@@ -10,7 +10,7 @@
 // Copyright (C) 2011 mod_csp project
 // License : Apache Licence 2.0
 //
-// version : 0.01
+// version : 0.0.3
 //
 // Last modification by kumatiki
 //
@@ -33,6 +33,9 @@
 #include <apr_strings.h>
 #include <unistd.h>
 #include <apr_time.h>
+#include <apr_signal.h>
+#include<setjmp.h>
+
 
 module AP_MODULE_DECLARE_DATA csp_module;
 
@@ -65,15 +68,6 @@ module AP_MODULE_DECLARE_DATA csp_module;
 ////////////////////////////////////////////////////////////////////////////////////////////////
 //		include
 ////////////////////////////////////////////////////////////////////////////////////////////////
-# 1 "include.c"
-# 1 "<built-in>"
-# 1 "<R}hC>"
-# 1 "include.c"
-
-
-
-
-
 
 # 1 "growmem.c" 1
 # 9 "growmem.c"
@@ -85,7 +79,7 @@ typedef struct _growMem {
 } GROWMEM;
 
 
-GROWMEM * getGrowMem(apr_pool_t *p,int size){
+static GROWMEM * getGrowMem(apr_pool_t *p,int size){
  if(size < 1024)size = 1024;
  GROWMEM * gm = apr_pcalloc(p,sizeof(GROWMEM));
  gm->p = p;
@@ -95,7 +89,7 @@ GROWMEM * getGrowMem(apr_pool_t *p,int size){
  return gm;
 }
 
-void gm_putc(GROWMEM * gm,char c){
+static void gm_putc(GROWMEM * gm,char c){
  if(gm->size < gm->index + 1){
   char * tmp = apr_pcalloc(gm->p,gm->size * 2);
   memcpy(tmp,gm->mem,gm->size);
@@ -106,7 +100,7 @@ void gm_putc(GROWMEM * gm,char c){
  gm->index++;
 }
 
-void gm_puts(GROWMEM * gm,char * str){
+static void gm_puts(GROWMEM * gm,char * str){
  int len = strlen(str);
 
  if(gm->size < gm->index + len){
@@ -123,14 +117,15 @@ void gm_puts(GROWMEM * gm,char * str){
  gm->index +=len;
 }
 
-char * gm_getmem(GROWMEM * gm){
+static char * gm_getmem(GROWMEM * gm){
  return gm->mem;
 }
 
-int gm_getsize(GROWMEM * gm){
+static int gm_getsize(GROWMEM * gm){
  return gm->index;
 }
 # 8 "include.c" 2
+
 # 1 "file.c" 1
 
 
@@ -138,7 +133,7 @@ int gm_getsize(GROWMEM * gm){
 
 
 
-char * fileRead(apr_pool_t * p,char * filename,int * s)
+static char * fileRead(apr_pool_t * p,char * filename,int * s)
 {
  apr_file_t *fpr;
     apr_file_open(&fpr, filename, APR_READ, APR_OS_DEFAULT , p);
@@ -155,6 +150,7 @@ char * fileRead(apr_pool_t * p,char * filename,int * s)
  return buffer;
 }
 # 9 "include.c" 2
+
 # 1 "param.c" 1
 # 10 "param.c"
 struct csp_config {
@@ -182,6 +178,7 @@ static const char *cmd_csp_message (cmd_parms *parms, void *mconfig, char *arg){
 
 
 typedef const char *(*CMD_HAND_TYPE) (cmd_parms*,void*);
+
 static const command_rec csp_cmds[] = {
   {
  "csp_tmp",
@@ -194,9 +191,10 @@ static const command_rec csp_cmds[] = {
  {0},
 };
 # 10 "include.c" 2
+
 # 1 "cookie.c" 1
 # 10 "cookie.c"
-void setCookie(request_rec *r,char * args1,char * args2){
+static void setCookie(request_rec *r,char * args1,char * args2){
 
  char * c1 = apr_pstrdup(r->pool, "Set-Cookie");
  char * c2 = apr_pstrdup(r->pool, apr_pstrcat(r->pool,args1,"=",args2,NULL));
@@ -207,7 +205,7 @@ void setCookie(request_rec *r,char * args1,char * args2){
 
 
 
-apr_table_t * getCOOKIE(request_rec* r) {
+static apr_table_t * getCOOKIE(request_rec* r) {
 
  const char * cookie = apr_table_get(r->headers_in, "Cookie");
     if (NULL == cookie)
@@ -234,9 +232,10 @@ apr_table_t * getCOOKIE(request_rec* r) {
     return cookei_params;
 }
 # 11 "include.c" 2
+
 # 1 "getpost.c" 1
 # 10 "getpost.c"
-apr_table_t * getGET(request_rec* r) {
+static apr_table_t * getGET(request_rec* r) {
 
     if (NULL == r->args)
         return NULL;
@@ -268,7 +267,7 @@ apr_table_t * getGET(request_rec* r) {
 
 
 
-apr_table_t * getPOST(request_rec* r) {
+static apr_table_t * getPOST(request_rec* r) {
     char* result = "";
     int total_bytes = 0;
 
@@ -276,7 +275,6 @@ apr_table_t * getPOST(request_rec* r) {
         return NULL;
     }
 
-
     if (ap_should_client_block(r)) {
         while (total_bytes<POSTMAX) {
             char buffer[BUFFLEN];
@@ -313,9 +311,10 @@ apr_table_t * getPOST(request_rec* r) {
     return NULL;
 }
 # 12 "include.c" 2
+
 # 1 "trans.c" 1
 # 10 "trans.c"
-int includetenkai(apr_pool_t * p,GROWMEM * buf,char * filename,char * dstdir,request_rec *r)
+static int includetenkai(apr_pool_t * p,GROWMEM * buf,char * filename,char * dstdir,request_rec *r)
 {
 
 
@@ -324,30 +323,38 @@ int includetenkai(apr_pool_t * p,GROWMEM * buf,char * filename,char * dstdir,req
 
  int size;
  char * buffer = fileRead(p,dstfile,&size);
- if(buffer==NULL)return CSPERROR;
+
+ if(buffer==NULL)
+  return CSPERROR;
 
  int i = 0;
  while(i<size){
-  if(strncmp((buffer+i),"<%include",9)==0){
+
+  if(strncmp((buffer+i),"<%include",9)==0 && strncmp((buffer+i),"<%includepath",13)!=0 ){
    char * tmp = buffer+i;
    char * tmp1 = strchr(tmp,'\"');
    if(tmp1==NULL)return CSPERROR;
    char * tmp2 = strchr(tmp1+1,'\"');
    if(tmp2==NULL)return CSPERROR;
    char * tmp3 = strstr(tmp2+1,"%>");
-   if(tmp3==NULL || tmp1-tmp1 > 1024 )return CSPERROR;
+
+   if(tmp3==NULL || tmp1-tmp1 > 1024 )
+    return CSPERROR;
 
    int len = tmp2-tmp1-1;
    char * nfile = apr_pcalloc(p,1024);
    strncpy(nfile,tmp1+1,len);
    int ret = includetenkai(p,buf,nfile,dstdir,r);
-   if(ret==CSPERROR)return CSPERROR;
+
+   if(ret==CSPERROR)
+    return CSPERROR;
 
    i+=tmp3-tmp+2;
   }
   gm_putc(buf,* (buffer+i));
   i++;
  }
+ return CSPOK;
 }
 
 
@@ -355,7 +362,7 @@ int includetenkai(apr_pool_t * p,GROWMEM * buf,char * filename,char * dstdir,req
 
 
 
-int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM * pinclude ,GROWMEM * plib,request_rec *r){
+static int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM * pinclude ,GROWMEM * plib,request_rec *r){
 
  int mode1flag = 0;
 
@@ -485,22 +492,61 @@ int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM *
  apr_file_open(&fpw, dstcfile, APR_WRITE|APR_CREATE|APR_TRUNCATE|APR_FOPEN_BINARY, APR_OS_DEFAULT , p);
 
  apr_file_puts(
+
 # 1 "head.h" 1
+"////////////////////////////////////////////////////////////////////////////////////////////////\n"
+"//\n"
+"//  mod_csp -- Apache mod_csp module\n"
+"//\n"
+"////////////////////////////////////////////////////////////////////////////////////////////////\n"
+"\n"
+"// for printf\n"
+"#include <stdio.h>\n"
+"\n"
 "#include <httpd.h>\n"
 "#include <http_config.h>\n"
 "#include <http_protocol.h>\n"
 "#include <ap_config.h>\n"
 "#include <apr_hash.h>\n"
 "#include <apr_strings.h>\n"
+"#include <signal.h>\n"
+"#include <setjmp.h>\n"
+"\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"//define \n"
+"///////////////////////////////////////////////////////////////////////\n"
+"\n"
+"#define GET		__get__\n"
+"#define POST	__post__\n"
+"#define COOKIE	__cookie__\n"
+"\n"
+"#define tbl_set(a,b,c)		apr_table_set(a,b,c)\n"
+"#define tbl_add(a,b,c)		apr_table_add(a,b,c)\n"
+"#define tbl_clr(a)			apr_table_clear(a);\n"
+"#define tbl_unset(a,b)		apr_table_unset(a,b)\n"
 "\n"
-"//macro\n"
 "#undef printf\n"
-"#define printf(arg,...) (arg==NULL)?:ap_rprintf(__r__,arg, ##__VA_ARGS__)\n"
-"#define strjoin(...) apr_pstrcat(__p__,__VA_ARGS__,NULL)\n"
+"#define printf(arg,...) (arg==NULL)?NONE:ap_rprintf(__r__,arg, ##__VA_ARGS__)\n"
+"#define psprintf(arg,...) (arg==NULL)?NONE:apr_psprintf(__r__,arg, ##__VA_ARGS__)\n"
+"#define stnjoin(...) apr_pstrcat(__p__,__VA_ARGS__,NULL)\n"
+"#define stncat(...) apr_pstrcat(__p__,__VA_ARGS__,NULL)\n"
+"#define stndup(a) apr_pstrdup (__p__,a)\n"
 "\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"//	typedef\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"typedef apr_table_t * table;\n"
+"\n"
+"\n"
+"///////////////////////////////////////////////////////////////////////\n"
 "//const\n"
+"///////////////////////////////////////////////////////////////////////\n"
 "const char * NONE = \"\";\n"
 "\n"
+"\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"//		struct\n"
+"///////////////////////////////////////////////////////////////////////\n"
 "//session\n"
 "struct _sessionTable {\n"
 "	int datasize;\n"
@@ -521,7 +567,9 @@ int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM *
 "};\n"
 "\n"
 "\n"
+"///////////////////////////////////////////////////////////////////////\n"
 "//\n"
+"///////////////////////////////////////////////////////////////////////\n"
 "static request_rec * __r__;\n"
 "static apr_pool_t * __p__;\n"
 "static apr_table_t * __get__;\n"
@@ -530,8 +578,85 @@ int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM *
 "static 	int * __session_index__;\n"
 "static struct _sessionTable * __session__;\n"
 "\n"
+"//setjump as core dump \n"
+"jmp_buf jbuf;\n"
 "\n"
-"//library function\n"
+"\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"//		hash\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"//#define hash	apr_hash_t * \n"
+"\n"
+"\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"//		table\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"apr_table_t * tbl_make(int size){\n"
+"	return apr_table_make(__p__, size);\n"
+"}\n"
+"\n"
+"int tbl_size(apr_table_t * __table__){\n"
+"	if( __table__==NULL)\n"
+"		return 0;\n"
+"	const apr_array_header_t * arr = apr_table_elts(__table__);\n"
+"	return arr->nelts;\n"
+"}\n"
+"\n"
+"char * tbl_key_i(apr_table_t * __table__,int i){\n"
+"	if(__table__==NULL)\n"
+"		return NONE;\n"
+"	const apr_array_header_t *arr = apr_table_elts(__table__);\n"
+"	int size = arr->nelts;\n"
+"\n"
+"	if(i >= size)\n"
+"		return NONE;\n"
+"	apr_table_entry_t *entry =	(apr_table_entry_t *)(arr->elts + (arr->elt_size * i));\n"
+"	return entry->key;\n"
+"}\n"
+"\n"
+"char * tbl_val_i(apr_table_t * __table__,int i){\n"
+"	if(__table__==NULL)\n"
+"		return NONE;\n"
+"	const apr_array_header_t *arr = apr_table_elts(__table__);\n"
+"	int size = arr->nelts;\n"
+"	if(i >= size)\n"
+"		return NONE;\n"
+"	apr_table_entry_t *entry =	(apr_table_entry_t *)(arr->elts + (arr->elt_size * i));\n"
+"	return entry->val;\n"
+"}\n"
+"\n"
+"char *  tbl_val(apr_table_t * __table__,char * args){\n"
+"	if(__table__==NULL)	return NONE;\n"
+"	if(args==NULL)	return NONE;\n"
+"\n"
+"	char * tmp = apr_table_get(__table__,args);\n"
+"	if(tmp==NULL)return NONE;\n"
+"	return tmp;\n"
+"}\n"
+"\n"
+"\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"//		string\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"char * stnreplace(char *buf, char *pre, char *aft)\n"
+"{\n"
+"	char * p;\n"
+"	p = strstr(buf, pre);\n"
+"	if (p){\n"
+"		char *p_2 = p + strlen(pre);\n"
+"		char * buf2 = stnreplace(p_2, pre, aft);\n"
+"		int s = p-buf;\n"
+"		buf = apr_pstrdup(__p__,buf);\n"
+"		* (buf + s) = 0;\n"
+"		return stnjoin(buf,aft,buf2);\n"
+"	}\n"
+"	return buf;\n"
+"}\n"
+"\n"
+"\n"
+"///////////////////////////////////////////////////////////////////////\n"
+"//		Exclusive use\n"
+"///////////////////////////////////////////////////////////////////////\n"
 "char *  getGet(char * args){\n"
 "	if(__get__==NULL)	return NONE;\n"
 "	if(args==NULL)	return NONE;\n"
@@ -552,7 +677,7 @@ int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM *
 "\n"
 "void setCookie(char * args1,char * args2){\n"
 "	char * c1 = apr_pstrdup(__p__, \"Set-Cookie\");\n"
-"	char * c2 = apr_pstrdup(__p__, strjoin(args1,\"=\",args2));\n"
+"	char * c2 = apr_pstrdup(__p__, stnjoin(args1,\"=\",args2));\n"
 "	apr_table_addn(__r__->headers_out, c1,c2);\n"
 "}\n"
 "\n"
@@ -582,7 +707,7 @@ int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM *
 "			return __session__[i].data;\n"
 "		}\n"
 "	}\n"
-"	return \"\";\n"
+"	return NONE;\n"
 "}\n"
 "\n"
 "void setSession_ptr(char * key,void * data,int size){\n"
@@ -614,6 +739,50 @@ int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM *
 "	return NULL;\n"
 "}\n"
 "\n"
+"\n"
+"//signal handler \n"
+"void sigint_handler(int s){\n"
+"\n"
+"	int i, nr;\n"
+"	void *trace[128];\n"
+"	char **ents;\n"
+"\n"
+"\n"
+"	switch(s){\n"
+"		case SIGBUS:\n"
+"	 		ap_rprintf(__r__, \"<br><font color=red size=6>Bus Error</font>\\n\");\n"
+"			break;\n"
+"		case SIGSEGV:\n"
+"	 		ap_rprintf(__r__, \"<br><font color=red size=6>Segmentation Violation</font>\\n\");\n"
+"			break;\n"
+"		case SIGALRM:\n"
+"	 		ap_rprintf(__r__, \"<br><font color=red size=6>CSP Time Out</font>\\n\");\n"
+"			break;\n"
+"		default:\n"
+"	 		ap_rprintf(__r__, \"<br><font color=red size=6>signal %d </font>\\n\",s);\n"
+"			break;\n"
+"	}\n"
+"\n"
+"	nr = backtrace(trace, 128);\n"
+"	if (nr>=128){\n"
+"	    printf(\"Stack too deep, some entries are stripped off\\n\");  \n"
+"	} else {\n"
+"		ents = backtrace_symbols(trace, nr);\n"
+"\n"
+"		if (ents == NULL) {\n"
+"			printf(\"backtrace_symbols !!!\");\n"
+"		} else {\n"
+"			printf(\"<hr>\");\n"
+"			for(i=0; i<nr; i++){\n"
+"				printf(\"#%d : %s<br>\\n\", i, ents[i]);\n"
+"			}\n"
+"		}\n"
+"	}\n"
+"\n"
+"\n"
+"	longjmp( jbuf, 1 );\n"
+"}\n"
+"\n"
 "//this function is called from apache\n"
 "int callmain( struct _param * param ){\n"
 "\n"
@@ -628,16 +797,25 @@ int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM *
 "	__session_index__ = param->session_index;\n"
 "	__session__ = param->session;\n"
 "\n"
-"	//ap_rprintf(__r__, \"session_index: %d<br>\\n\", * __session_index__);\n"
-"	//ap_rprintf(__r__, \"func: %lx<br>\\n\", param->func);\n"
-"	//p_rprintf(__r__, \"func: %d<br>\\n\", (* param->func)());\n"
+"\n"
+"	signal( SIGSEGV, &sigint_handler );\n"
+"	signal( SIGTSTP , &sigint_handler );\n"
+"	signal( SIGILL , &sigint_handler );\n"
+"	signal( SIGHUP , &sigint_handler );\n"
+"	signal( SIGFPE  , &sigint_handler );\n"
+"	signal( SIGALRM  , &sigint_handler );\n"
+"	alarm(5);\n"
+"\n"
+"	if( setjmp( jbuf ) == 1 ) {\n"
+"		return 1;\n"
+"	}\n"
+"\n"
 "	main();\n"
+"\n"
 "	return 0;\n"
 "}\n"
 "\n"
-"\n"
-"\n"
-# 181 "trans.c" 2
+# 189 "trans.c" 2
  ,fpw);
 
  char * t11 = gm_getmem(t1);
@@ -667,6 +845,7 @@ int csptoc(apr_pool_t *p, GROWMEM * ten,char * dstcfile,char * dstdir,GROWMEM *
  return 0;
 }
 # 13 "include.c" 2
+
 # 1 "session.c" 1
 # 10 "session.c"
 struct _sessionTable {
@@ -687,7 +866,7 @@ struct _param {
 };
 
 
-char * makeSessionID(request_rec *r){
+static char * makeSessionID(request_rec *r){
  char * sessionID = apr_psprintf(r->pool,"ID_%lld",apr_time_now());
 }
 
@@ -695,7 +874,7 @@ char * makeSessionID(request_rec *r){
 
 
 
-void sessionGet(int * session_index,struct _sessionTable * session,char * sessionid,apr_pool_t * p,request_rec *r){
+static void sessionGet(int * session_index,struct _sessionTable * session,char * sessionid,apr_pool_t * p,request_rec *r){
 
     char * sesdir;
  int i;
@@ -734,7 +913,7 @@ void sessionGet(int * session_index,struct _sessionTable * session,char * sessio
 
 
 
-void sessionSet(int * session_index,struct _sessionTable * session,char * sessionid,apr_pool_t * p,request_rec *r){
+static void sessionSet(int * session_index,struct _sessionTable * session,char * sessionid,apr_pool_t * p,request_rec *r){
     char * sesdir;
  int i;
 
@@ -766,14 +945,12 @@ void sessionSet(int * session_index,struct _sessionTable * session,char * sessio
  }
  apr_file_close(fpw);
 }
-# 14 "include.c" 2
-
 
 
 ////////////////////////////////////////////////////////////////////////////////////////////////
 //		debug
 ////////////////////////////////////////////////////////////////////////////////////////////////
-void print_table(request_rec *r,apr_table_t *table) {
+static void print_table(request_rec *r,apr_table_t *table) {
 	if (apr_is_empty_table(table)) {
 		ap_rprintf(r,"There is no element. <br>\n");
 	}
@@ -787,6 +964,17 @@ void print_table(request_rec *r,apr_table_t *table) {
 	}
 }
 
+
+//static void my_signal_handler (int x) {
+
+void errordispay(request_rec *r,char * title,char *contents){
+	ap_rprintf(r,"<h2>%s</h2>\n",title);
+	ap_rprintf(r,"<hr>\n<pre>\n");
+	ap_rprintf(r,"%s",contents);
+	ap_rprintf(r,"</pre>\n");
+}
+
+
 ////////////////////////////////////////////////////////////////////////////////////////////////
 //
 //		handler
@@ -807,6 +995,10 @@ static int csp_handler(request_rec *r)
 	    return OK;
 
 
+//	apr_signal(SIGINT, my_signal_handler);
+	// Error buffer
+//    GROWMEM * errbuffer = getGrowMem(p,1024);
+
 	// make folder name by accessed file 
 	char * folder = apr_pstrdup(p,(const char * )r->filename);
 	tmp = strrchr(folder,'/');
@@ -869,6 +1061,7 @@ static int csp_handler(request_rec *r)
 	mtime2 = finfo2.ctime;
 	apr_file_close(fpr);
 
+
 	//It copies it when there is an update. 
 	if(mtime2>mtime1){
 
@@ -883,6 +1076,7 @@ static int csp_handler(request_rec *r)
 			return OK;
 		}
 
+
 	   //trans
 	    GROWMEM * binclude = getGrowMem(p,1024);
     	GROWMEM * blib = getGrowMem(p,1024);
@@ -897,14 +1091,20 @@ static int csp_handler(request_rec *r)
 		    apr_file_open(&fpr, cat(dstfile,".log"), APR_READ, APR_OS_DEFAULT , p);
 			apr_file_read(fpr,errorlog,&len);
 			apr_file_close(fpr);
-
-			ap_rprintf(r,"<h2>compile error </h2>");
-			ap_rprintf(r,"<hr><pre>");
-			ap_rprintf(r,"%s",errorlog);
-			ap_rprintf(r,"</pre>");
+			errordispay(r,"compile error",errorlog);
 			return   OK;
 		}
-		system(cat("gcc -shared -Wl,-soname,lib",filename,".so -o ",dstdir,"lib",filename,".so  ",dstofile," ",gm_getmem(blib)));
+		system(cat("gcc -shared -Wl,-soname,lib",filename,".so -o ",dstdir,"lib",filename,".so  ",dstofile," ",gm_getmem(blib)," 2>",dstfile,"1.log"));
+		if(ccc!=0){
+  			char * errorlog = (char * )apr_pcalloc(p,1024*10);
+  			int len = 1024*10;
+		    apr_file_open(&fpr, cat(dstfile,"1.log"), APR_READ, APR_OS_DEFAULT , p);
+			apr_file_read(fpr,errorlog,&len);
+			apr_file_close(fpr);
+			errordispay(r,"link error",errorlog);
+			return   OK;
+		}
+
 	}
 
 	//go go go
@@ -926,8 +1126,13 @@ static int csp_handler(request_rec *r)
 			st = (* fn)(&param);
 	   }
 	   apr_dso_unload(dso_h);
-	} else
-		return   HTTP_INTERNAL_SERVER_ERROR;
+	} else {
+		char errstr[1024];
+		apr_dso_error(dso_h, errstr, 1024);
+		errordispay(r,"dso error",errstr);
+		return OK;
+//		return   HTTP_INTERNAL_SERVER_ERROR;
+	}
 
 //	save session
 	if(session_index!=0){
